Timer  = require 'hump.timer'
class  = require 'hump.class'
GS     = require 'hump.gamestate'
vector = require 'hump.vector'
Camera = require 'hump.camera'
local signals = require 'signals'

-- helper stuff
function head(a, ...) return a end
function tail(a, ...) return ... end
function COLOR(hex)
	local r,g,b,a = hex:match('^#-(%x%x)(%x%x)(%x%x)(%x-)$')
	return tonumber(r,16), tonumber(g,16), tonumber(b,16), a and tonumber(a,16)
end

function Serialize(t)
	local r = {"{"}
	for k, v in pairs(t) do
		if type(k) == "number" then
			k = "["..k.."]"
		elseif type(k) == "string" then
			k = ("[%q]"):format(k)
		end

		local sep = ','
		local valtype = type(v)
		if valtype == "table" then
			v = Serialize(v)
			sep = ',\n'
		elseif valtype == "string" then
			v = ('%q'):format(v)
		elseif valtype == "number" then
			-- check for +-inf and nan
			v = tostring(v):gsub('inf', 'math.huge'):gsub('nan', '0/0')
		else
			v = tostring(v)
		end
		r[#r+1] = k
		r[#r+1] = '='
		r[#r+1] = v
		r[#r+1] = sep
	end
	r[#r+1] = '}'
	return table.concat(r)
end

DT = love.timer.getDelta
_NULL_ = function() end

--[[ RESOURCE MANAGEMENT ]]--
local function Proxy(f)
	return setmetatable({}, {__index = function(self, k)
		local v = f(k)
		rawset(self, k, v)
		return v
	end})
end

--State = Proxy(function(k) return assert(love.filesystem.load('states/'..k..'.lua'))() end)
State = Proxy(function(k) return require('states.'..k) end)
Font = {
	fancy  = Proxy(function(k) return love.graphics.newFont('font/Homev2.ttf', k) end),
	normal = Proxy(function(k) return love.graphics.newFont('font/aur.ttf', k) end),
	debug  = Proxy(function(k) return love.graphics.newFont(k) end),

}
Image  = Proxy(function(k) return love.graphics.newImage('img/' .. k .. '.png') end)
Sound  = Proxy(function(k) return love.audio.newSource('sound/' .. k .. '.ogg') end)
Stream = Proxy(function(k) return love.audio.newSource('sound/' .. k .. '.ogg', 'stream') end)
Map    = Proxy(function(k) return assert(love.filesystem.load('maps/'..k..'.lua'))() end)

MapList = {next = {}, number = {}}
for _,path in ipairs(love.filesystem.enumerate('maps')) do
	MapList[#MapList+1] = path:match('^(.-)%.lua$')
end
table.sort(MapList, function(a,b)
	local an,bn = tonumber(a), tonumber(b)
	if an and bn then return an < bn end
	if an and not bn then return true end
	if bn and not an then return false end
	return a < b
end)

for i = 1,#MapList do
	MapList.next[MapList[i]] = MapList[i+1]
	MapList.number[MapList[i]] = i
end

local ok
ok, Config = pcall(love.filesystem.load, 'config')
if not ok then
	Config = {
		done = 0,
		hide_dialog = {}
	}
else
	Config = Config()
end
Config.done = Config.done or 0

--[[ LESS ANNOYING COLOR SETTING ]]--
do
	local setColor = love.graphics.setColor
	local setBG = love.graphics.setBackgroundColor
	function love.graphics.setColor(r,g,b,a)
		if type(r) == 'table' then
			r,g,b,a = unpack(r)
		end
		if r <= 1 and g <= 1 and b <= 1 then
			r,g,b,a = 255*r,255*g,255*b,255*(a or 1)
		end
		setColor(r,g,b,a)
	end
	function love.graphics.setBackgroundColor(r,g,b,a)
		if type(r) == 'table' then
			r,g,b,a = unpack(r)
		end
		if r <= 1 and g <= 1 and b <= 1 then
			r,g,b,a = 255*r,255*g,255*b,255*(a or 1)
		end
		setBG(r,g,b,a)
	end
end

--[[ Main program ]]--
function love.load()
	love.keyboard.setKeyRepeat(.25)
	math.randomseed(os.time()) math.random()
	WIDTH  = love.graphics.getWidth()
	HEIGHT = love.graphics.getHeight()

	-- audio
	Stream.music:setVolume(.2)
	Stream.music:setLooping(true)
	if not Config.no_music then
		Stream.music:play()
	end

	local function invalid() Sound.invalid:play() end
	--signals.register('invalid-operation', invalid)
	signals.register('connection-invalid', invalid)
	signals.register('pickup-success', function() Sound.pickup:play() end)
	signals.register('drop-success', function() Sound.drop:play() end)
	signals.register('circuit-step', function() Sound.step:play() end)


	GS.registerEvents()
	GS.switch(State.title)
	--GS.switch(State.levelselect)
	--GS.switch(State.game)
	--GS.switch(State.editor)
end

function love.update(dt)
	Timer.update(dt)
end

function love.quit()
	local s = Serialize(Config)
	local f = love.filesystem.newFile('config')
	f:open('w')
	f:write('return ')
	f:write(s)
	f:close()

	-- change to title screen to make the editor save the current progress
	GS.switch(State.title)
end


-- overpovered pause screen
if love.graphics.isSupported('canvas', 'pixeleffect') then

local canvas
local blur = love.graphics.newPixelEffect[[
extern vec2 dim;
vec4 effect(vec4 c, Image tex, vec2 tc, vec2 q)
{
	vec2 o = vec2(1.)/dim;
	c  = Texel(tex, tc) * 3.;
	c += Texel(tex, tc-vec2( o.x,.0)) * 2.;
	c += Texel(tex, tc-vec2(-o.x,.0)) * 2.;

	c += Texel(tex, tc-vec2( o.x, o.y)) * 1.;
	c += Texel(tex, tc-vec2(  .0, o.y)) * 2.;
	c += Texel(tex, tc-vec2(-o.x, o.y)) * 1.;

	c += Texel(tex, tc-vec2( o.x,-o.y)) * 1.;
	c += Texel(tex, tc-vec2(  .0,-o.y)) * 2.;
	c += Texel(tex, tc-vec2(-o.x,-o.y)) * 1.;

	return c / 15.5;
}
]]
function love.keypressed(key)
	if key == 'escape' then
		local r = {}
		for _,cb in ipairs{'update', 'draw', 'keypressed', 'keyreleased', 'mousepressed', 'mousereleased'} do
			r[cb] = love[cb]
			love[cb] = _NULL_
		end

		canvas = canvas or love.graphics.newCanvas()
		canvas:clear()
		blur:send('dim', {WIDTH, HEIGHT})
		love.graphics.setCanvas(canvas)
		r.update(0)
		r.draw()
		love.graphics.setPixelEffect(blur)
		for i = 1,15 do
			love.graphics.draw(canvas,0,0)
		end
		love.graphics.setPixelEffect()
		love.graphics.setCanvas()

		local intitle = GS.current() == State.title

		function love.draw()
			love.graphics.setColor(1,1,1)
			love.graphics.draw(canvas,0,0)
			local s = 'PAUSE'
			if intitle then s = 'Quit?' end
			love.graphics.setFont(Font.normal[30])
			local h = Font.normal[60]:getHeight(s)
			love.graphics.print(s, (WIDTH-Font.normal[30]:getWidth(s))/2, (HEIGHT-h)/2)

			s = '(escape to go to title)'
			if intitle then s = '(escape to confirm)' end
			love.graphics.setFont(Font.normal[20])
			love.graphics.print(s, (WIDTH-Font.normal[20]:getWidth(s))/2, (HEIGHT-Font.normal[20]:getHeight(s)+h)/2)
		end

		function love.keypressed(key)
			if key == 'escape' then
				if intitle then love.event.push('quit') end
				GS.switch(State.title)
			end
			for _,cb in ipairs{'update', 'draw', 'keypressed', 'keyreleased', 'mousepressed', 'mousereleased'} do
				love[cb] = r[cb]
			end
		end
	end
end

else -- your hardware sucks

function love.keypressed(key)
	if key == 'escape' then
		local r = {}
		for _,cb in ipairs{'update', 'draw', 'keypressed', 'keyreleased', 'mousepressed', 'mousereleased'} do
			r[cb] = love[cb]
			love[cb] = _NULL_
		end

		local intitle = GS.current() == State.title

		function love.draw()
			r.draw()
			love.graphics.setColor(0,0,0,.8)
			love.graphics.rectangle('fill', 0,0, WIDTH, HEIGHT)

			love.graphics.setColor(1,1,1)
			local s = 'PAUSE'
			if intitle then s = 'Quit?' end
			love.graphics.setFont(Font.normal[30])
			local h = Font.normal[60]:getHeight(s)
			love.graphics.print(s, (WIDTH-Font.normal[30]:getWidth(s))/2, (HEIGHT-h)/2)

			s = '(escape to go to title)'
			if intitle then s = '(escape to confirm)' end
			love.graphics.setFont(Font.normal[20])
			love.graphics.print(s, (WIDTH-Font.normal[20]:getWidth(s))/2, (HEIGHT-Font.normal[20]:getHeight(s)+h)/2)
		end

		function love.update(dt) r.update(0) end

		function love.keypressed(key)
			if key == 'escape' then
				if intitle then love.event.push('quit') end
				GS.switch(State.title)
			end
			for _,cb in ipairs{'update', 'draw', 'keypressed', 'keyreleased', 'mousepressed', 'mousereleased'} do
				love[cb] = r[cb]
			end
		end
	end
end

end
