local st = GS.new()
local signals = require 'signals'
local gui = require 'quickie'
require 'circuit-editor'
require 'level'

function st:init()
	self.editor = CircuitEditor()
	self.show_editor = {checked = true}
	self.run = {checked = false}
	self.time = 0
	self.speed = {value = 1, min = .1, max = 2}
	self.realdraw = self.draw
end

function st:enter(_, level, from_editor)
	if not level then GS.switch(State.title) return end
	love.graphics.setFont(Font.normal[10])
	self.level = Level(Map[level])
	self.from_editor = from_editor
	local changing_level = false
	self.slot = signals.register('level-complete', function()
		if changing_level then return end
		changing_level = true

		Sound.level_done:play()
		self.run.checked = false
		if from_editor then
			Timer.add(.1, function() GS.switch(State.editor) end)
		else
			-- fade out
			local fade = Timer.Interpolator(1.3, function(f) return (1-math.cos(f*math.pi))/2 end)
			function st:draw()
				self:realdraw()
				local a = fade(DT()) or 1
				love.graphics.setColor(0,0,0, a)
				love.graphics.rectangle('fill', 0,0, WIDTH, HEIGHT)
			end
			Timer.add(1.3, function()
				self.draw = self.realdraw
				GS.switch(State.game, MapList.next[level])
			end)

			Config.done = math.max(Config.done, MapList.number[level])
		end
	end)

	-- fade in
	local fade = Timer.Interpolator(.5, function(f) return (1+math.cos(f*math.pi))/2 end)
	function st:draw()
		self:realdraw()
		local a = fade(DT()) or 0
		love.graphics.setColor(0,0,0, a)
		love.graphics.rectangle('fill', 0,0, WIDTH, HEIGHT)
	end
	Timer.add(.5, function() self.draw = self.realdraw end)

	if not Config.hide_dialog[MapList.number[level]] then
		self.dialog = Map[level].dialog
	else
		self.dialog = nil
	end
end

function st:leave()
	if self.slot then
		signals.remove('level-complete', self.slot)
	end
end

function st:update(dt)
	if self.dialog then
		self.dialog(self, gui, 100,75,WIDTH-200,HEIGHT-150)
		return
	end

	self.level:update(dt, self.speed.value)

	love.graphics.setFont(Font.normal[10])
	local x,y = WIDTH-5, 5
	x = x -  64;
	if self.show_editor.checked then
		if gui.Button('clear', x,y,    64,22) then
			self.editor.circuit = Circuit()
			self.run.checked = false
		end
		y = y + 27
	end
	if self.from_editor then
		if gui.Button('edit', x,y,    64,22) then
			GS.switch(State.editor)
		end
		y = y + 27
	end
	y = 5
	x = x - 100; gui.Label('show circuit',      x,y+3, nil,nil,'left')
	x = x -  16; gui.Checkbox(self.show_editor, x,y+6,  12,12)
	x = x -  76; gui.Label('simulate',          x,y+3, nil,nil,'left')
	x = x -  16; if gui.Checkbox(self.run,      x,y+6,  12,12) and self.run.checked then
		self.level:reset()
		self.editor.circuit:clear()
	end
	x = x - 100; gui.Slider(self.speed,         x,y+4,  80,17)
	x = x -  50; gui.Label('speed',             x,y+3, nil,nil,'left')

	if self.show_editor.checked then
		self.editor:update(dt)
	end

	if self.run.checked then
		self.time = self.time + dt
		if self.time > 1 / self.speed.value then
			self.time = 0
			self.editor.circuit:step()
		end
	end
end

function st:mousereleased(x,y, btn)
	if self.show_editor.checked then
		self.editor:mousereleased(x,y,btn)
	end
end

local glow = Timer.Oscillator(5,function(f) return (1-math.cos(2*math.pi*f))/2 end)
function st:draw()
	love.graphics.setColor(1,1,1, glow(DT())^5 * .3 + .7)
	love.graphics.draw(Image.bg, 0,0)
	self.level:draw()
	if self.show_editor.checked then
		self.editor:draw()
	end
	love.graphics.setLine(1)

	if self.dialog then
		love.graphics.setColor(.02,.03,.06,.9)
		love.graphics.rectangle('fill', 100,75, WIDTH-200,HEIGHT-150)
		love.graphics.setColor(.4,.4,.4,.5)
		love.graphics.rectangle('line', 100,75, WIDTH-200,HEIGHT-150)
	end

	gui.core.draw()
end

function st:keypressed(key)
	if key == 'e' then
		self.show_editor.checked = not self.show_editor.checked
	elseif key == ' ' then
		self.run.checked = not self.run.checked
		self.level:reset()
		self.editor.circuit:clear()
	elseif self.show_editor.checked then
		self.editor:keypressed(key)
	end
end

function st:keyreleased(key)
	if self.show_editor.checked then
		self.editor:keyreleased(key)
	end
end

return st
